#ifndef OBJ_H
#define OBJ_H

#include "aliases.h"
#include "img.h"
#include "mem_arena.h"
#include "typed_list.h"

#define INVALID_MODEL ((Model){0})
#define IS_INVALID_MODEL(m) (m.vertices == NULL || m.triangles == NULL)

typedef struct triangle Triangle;
struct triangle {
  u64 p0;
  u64 p1;
  u64 p2;
  u64 n0;
  u64 n1;
  u64 n2;
  u64 tx0;
  u64 tx1;
  u64 tx2;
};

typedef struct i64x2 V2i;
struct i64x2 {
  i64 x;
  i64 y;
};

typedef struct u64x2 V2u;
struct u64x2 {
  u64 x;
  u64 y;
};

typedef struct f32x2 V2f;
struct f32x2 {
  union {
    f32 x;
    f32 u;
  };
  union {
    f32 y;
    f32 v;
  };
};

typedef struct f32x3 V3f;
struct f32x3 {
  f32 x;
  f32 y;
  f32 z;
};

typedef struct f32x4 V4f;
struct f32x4 {
  f32 x;
  f32 y;
  f32 z;
  f32 w;
};

typedef struct u64x3 V3u;
struct u64x3 {
  u64 x;
  u64 y;
  u64 z;
};

typedef struct f32_3x3 M3x3f;
struct f32_3x3 {
  V3f row0;
  V3f row1;
  V3f row2;
};

typedef struct f32_4x4 M4x4f;
struct f32_4x4 {
  V4f row0;
  V4f row1;
  V4f row2;
  V4f row3;
};

typedef enum {
  RENDER_TYPE_WIREFRAME,
  RENDER_TYPE_FILLED,
  RENDER_TYPE_SHADED,

  COUNT_RENDER_TYPES,
} RenderType;

typedef enum {
  COLOUR_TYPE_FIXED,
  COLOUR_TYPE_RANDOM,

  COUNT_COLOUR_TYPE,
} ColourType;

typedef enum {
  PROJECTION_TYPE_ORTHOGRAPHIC,
  PROJECTION_TYPE_PERSPECTIVE,

  COUNT_PROJECTION_TYPE,
} ProjectionType;

MAKE_LIST_TYPE(V3f);
MAKE_LIST_TYPE(V2f);
MAKE_LIST_TYPE(Triangle);

typedef struct model Model;
struct model {
  LIST_TYPE(V3f) * vertices;
  LIST_TYPE(V3f) * normals;
  LIST_TYPE(V2f) * texture_coordinates;
  LIST_TYPE(Triangle) * triangles;
  Image *texture;
};

typedef struct render Render;
struct render {
  Image img;
  Depth depth;
};

Model load_obj_file(Arena *arena, const char *filename, const char *texture);
bool init_render(Arena *arena, Render *render, u64 width, u64 height);
void render_model(const Model *model, Render *render, Colour colour,
                  RenderType type, ColourType colour_type,
                  ProjectionType projection);

#endif // OBJ_H
